/*
  This file is part of GerQUtils.

  GerQUtils is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerQUtils is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerQUtils.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "associator.h"

#include "defaultassociator.h"
#include "sqlassoadapter.h"


namespace GerQ {

  QHash<QString, Associator*> Associator::fDefaultAssociators;

  Associator::Associator(
    AssoAdapter *aAdapter, DataContainer *aCtrYin, DataContainer *aCtrYan
  ) {
    fAdapter = aAdapter;
    fCtrYin = aCtrYin;
    fCtrYan = aCtrYan;
    fCtrYin -> addAssociator(this);
    fCtrYan -> addAssociator(this);
  } // Associator::Associator

  Associator::~Associator() {
    fCtrYin -> removeAssociator(this);
    fCtrYan -> removeAssociator(this);
  } // Associator::~Associator

  Associator* Associator::addDefaultAssociator(
    const QString &aTableName,
    const QString &aYinColumn, const QString &aYanColumn,
    DataContainer *aCtrYin, DataContainer *aCtrYan,
    void (*aAssociate)(DataObject*, DataObject*),
    void (*aDissociate)(DataObject*, DataObject*),
    void (*aGetYanData)(DataObject*, AssoData&),
    void (*aGetYinData)(DataObject*, AssoData&)
  ) {
    SqlAssoAdapter *adapter =
      new SqlAssoAdapter(aTableName, aYinColumn, aYanColumn);
    Associator *result = new DefaultAssociator(
      adapter, aCtrYin, aCtrYan,
      aAssociate, aDissociate, aGetYanData, aGetYinData);
    fDefaultAssociators.insert(aTableName, result);
    return result;
  } // Associator::addDefaultAssociator

  AssoAdapter* Associator::getAdapter() {
    return fAdapter;
  } // Associator::getAdapter

  Associator* Associator::getDefaultAssociator(const QString &aTableName) {
    if (fDefaultAssociators.contains(aTableName)) {
      return fDefaultAssociators.value(aTableName);
    } else {
      return NULL;
    } // if
  } // Associator::getDefaultAssociator

  void Associator::read(DataObject *aData, DataContainer *aContainer) {
    if (aContainer == fCtrYin) {
      readYin(aData);
    } else if (aContainer == fCtrYan) {
      readYan(aData);
    } // if
  } // Associator::read

  void Associator::readYan(DataObject *aYan) {
    AssoIDs IDs;
    AssoData data;
    fAdapter -> getYinIDs(aYan, IDs);
    getYinData(aYan, data);
    AssoDataIterator iData = data.begin();
    while (iData != data.end()) {
      DataObject *yin = *iData;
      QString ID = yin -> getID().toString();
      if (IDs.contains(ID)) {
        IDs.remove(ID);
      } else {
        dissociate(yin, aYan);
      } // if
      iData++;
    } // while
    AssoIDIterator iID = IDs.begin();
    while (iID != IDs.end()) {
      QString ID = *iID;
      associate(fCtrYin -> getData(ID), aYan);
      iID++;
    } // while
  } // Associator::readYan

  void Associator::readYin(DataObject *aYin) {
    AssoIDs IDs;
    AssoData data;
    fAdapter -> getYanIDs(aYin, IDs);
    getYanData(aYin, data);
    AssoDataIterator iData = data.begin();
    while (iData != data.end()) {
      DataObject *yan = *iData;
      QString ID = yan -> getID().toString();
      if (IDs.contains(ID)) {
        IDs.remove(ID);
      } else {
        dissociate(aYin, yan);
      } // if
      iData++;
    } // while
    AssoIDIterator iID = IDs.begin();
    while (iID != IDs.end()) {
      QString ID = *iID;
      associate(aYin, fCtrYan -> getData(ID));
      iID++;
    } // while
  } // Associator::readYin

  void Associator::write(DataObject *aData, DataContainer *aContainer) {
    if (aContainer == fCtrYin) {
      writeYin(aData);
    } else if (aContainer == fCtrYan) {
      writeYan(aData);
    } // if
  } // Associator::write

  void Associator::writeYan(DataObject *aYan) {
    AssoIDs IDs;
    AssoData data;
    fAdapter -> getYinIDs(aYan, IDs);
    getYinData(aYan, data);
    AssoDataIterator iData = data.begin();
    while (iData != data.end()) {
      DataObject *yin = *iData;
      QString ID = yin -> getID().toString();
      if (IDs.contains(ID)) {
        IDs.remove(ID);
      } else {
        fAdapter -> insertAssociation(yin, aYan);
      } // if
      iData++;
    } // while
    AssoIDIterator iID = IDs.begin();
    while (iID != IDs.end()) {
      QString ID = *iID;
      fAdapter -> deleteAssociation(fCtrYin -> getData(ID), aYan);
      iID++;
    } // while
  } // Associator::writeYan

  void Associator::writeYin(DataObject *aYin) {
    AssoIDs IDs;
    AssoData data;
    fAdapter -> getYanIDs(aYin, IDs);
    getYanData(aYin, data);
    AssoDataIterator iData = data.begin();
    while (iData != data.end()) {
      DataObject *yan = *iData;
      QString ID = yan -> getID().toString();
      if (IDs.contains(ID)) {
        IDs.remove(ID);
      } else {
        fAdapter -> insertAssociation(aYin, yan);
      } // if
      iData++;
    } // while
    AssoIDIterator iID = IDs.begin();
    while (iID != IDs.end()) {
      QString ID = *iID;
      fAdapter -> deleteAssociation(aYin, fCtrYan -> getData(ID));
      iID++;
    } // while
  } // Associator::writeYin

} // GerQ
