/*
  This file is part of GerQUtils.

  GerQUtils is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerQUtils is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerQUtils.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "datacontainer.h"

#include "associator.h"
#include "dataobject.h"
#include "defaultinstcontainer.h"
#include "defaultrefcontainer.h"
#include "defaultsqldataadapter.h"
#include "refcontainer.h"


namespace GerQ {

  QHash<QString, DataContainer*> DataContainer::fDefaultContainers;

  QVariant DataContainer::fDefaultNullID = QVariant();

  DataContainer::DataContainer(DataAdapter *aAdapter) {
    fAdapter = aAdapter;
    fAdapter -> fContainer = this;
    fParent = NULL;
    fSelf = NULL;
  } // DataContainer::DataContainer

  DataContainer::~DataContainer() {
    if (fSelf) {
      *fSelf = NULL;
    } // if
  } // DataContainer::~DataContainer

  void DataContainer::addAssociator(Associator* aAssociator) {
    if (!fAssociators.contains(aAssociator)) {
      fAssociators.insert(aAssociator, aAssociator);
    } // if
  } // DataContainer::addAssociator

  InstContainer* DataContainer::addInstDefault(
    const QString &aTableName, const QString &aIDColumn,
    DataObject* (*aCreateData)(),
    void (*aRead)(DataObject*, SqlDataAdapter*),
    void (*aWrite)(DataObject*, SqlDataAdapter*),
    RefContainer *aParent
  ) {
    /* Create adapter and container with the parameters given: */
    DefaultSqlDataAdapter *adapter = new DefaultSqlDataAdapter(
      aTableName, aIDColumn, aRead, aWrite);
    DefaultInstContainer *container = new DefaultInstContainer(
      adapter, aCreateData);
    container -> setParent(aParent);
    /* Store the container in the hash map for the default containers: */
    fDefaultContainers.insert(aTableName, container);
    return container;
  } // DataContainer::addInstDefault

  RefContainer* DataContainer::addRefDefault(
    const QString &aTableName, const QString &aIDColumn,
    void (*aRead)(DataObject*, SqlDataAdapter*),
    void (*aWrite)(DataObject*, SqlDataAdapter*),
    RefContainer *aParent
  ) {
    /* Create adapter and container with the parameters given: */
    DefaultSqlDataAdapter *adapter = new DefaultSqlDataAdapter(
      aTableName, aIDColumn, aRead, aWrite);
    DefaultRefContainer *container = new DefaultRefContainer(adapter);
    container -> setParent(aParent);
    /* Store the container in the hash map for the default containers: */
    fDefaultContainers.insert(aTableName, container);
    return container;
  } // DataContainer::addRefDefault

  DataIterator DataContainer::beginData() {
    return fData.begin();
  } // DataContainer::beginData

  void DataContainer::clearDefaultContainers() {
    while (fDefaultContainers.count()) {
      DataContainer *container = *(fDefaultContainers.begin() + 0);
      delete container;
    } // while
  } // DataContainer::clearDefaultContainers

  DataIterator DataContainer::endData() {
    return fData.end();
  } // DataContainer::endData

  DataAdapter* DataContainer::getAdapter() {
    return fAdapter;
  } // DataContainer::getAdapter

  void DataContainer::getAll() {
    QList<QVariant> IDs;
    fAdapter -> readAllIDs(IDs);
    QListIterator<QVariant> iter(IDs);
    while (iter.hasNext()) {
      DataObject *data = getData(iter.next());
      data -> read();
    } // while
  } // DataContainer::getAll

  int DataContainer::getCount() {
    return fData.count();
  } // DataContainer::getCount

  DataObject* DataContainer::getData(QVariant aID) {
    DataObject *result = NULL;
    QString ID = aID.toString();
    if (fData.contains(ID)) {
      /*
        There's already data present for the ID requested. Return it and be
	happy:
      */
      result = fData.value(ID);
      return result;
    } else {
      /* No data for the ID requested has been loaded yet. Try to do so now: */
      result = yetUnknownData(aID);
      /*
	If the adapter provided valid data for the ID requested, store for
	later use; if not, just the initial NULL will be returned:
      */
      if (result) {
	fData.insert(ID, result);
      } // if
      return result;
    } // if
  } // DataContainer::getData

  DataObject* DataContainer::getDataDefault(
    DataContainer *aContainer, SqlDataAdapter *aSql, const QString aFieldName
  ) {
    return aContainer -> getData(aSql -> getField(aFieldName));
  } // DataContainer::getDataDefault

  DataContainer* DataContainer::getDefaultContainer(QString &aTableName) {
    if (fDefaultContainers.contains(aTableName)) {
      return fDefaultContainers.value(aTableName);
    } else {
      return NULL;
    } // if
  } // DataContainer::getDefaultContainer

  QVariant DataContainer::getDefaultNullID() {
    return fDefaultNullID;
  } // DataContainer::getDefaultNullID

  QVariant DataContainer::getNullID() {
    return getDefaultNullID();
  } // DataContainer::getNullID

  RefContainer* DataContainer::getParent() {
    return fParent;
  } // DataContainer::getParent

  void* DataContainer::getSelf() {
    return fSelf;
  } // DataContainer::getSelf

  bool DataContainer::read(DataObject *aData) {
    bool present = fAdapter -> read(aData);
    if (present && fParent) {
      present = fParent -> read(aData);
    } // if
    if (present) {
      /* Avoid recursion: */
      QString ID = aData -> getID().toString();
      if (!fLoadInProgress.contains(ID)) {
        fLoadInProgress.insert(ID, ID);
	try {
	  QHashIterator<Associator*, Associator*> iter(fAssociators);
	  while (iter.hasNext()) {
	    iter.next();
	    iter.value() -> read(aData, this);
	  } // while
	FINALLY
	  fLoadInProgress.remove(ID);
	} // try
      } // if
    } // if
    return present;
  } // DataContainer::read

  void DataContainer::removeAssociator(Associator* aAssociator) {
    if (fAssociators.contains(aAssociator)) {
      fAssociators.remove(aAssociator);
    } // if
  } // DataContainer::removeAssociator

  void DataContainer::removeDefaultContainer(const QString &aTableName) {
    fDefaultContainers.remove(aTableName);
  } // DataContainer::removeDefaultContainer

  void DataContainer::setDataDefault(
    SqlDataAdapter *aSql, const QString aFieldName, DataObject *aData
  ) {
    if (aData) {
      aSql -> setField(aFieldName, aData -> getID());
    } else {
      aSql -> setField(aFieldName, aSql -> getContainer() -> getNullID());
    } // if
  } // DataContainer::setDataDefault

  void DataContainer::setDefaultNullID(const QVariant &aValue) {
    fDefaultNullID = aValue;
  } // DataContainer::setDefaultNullID

  void DataContainer::setParent(RefContainer *aValue) {
    /*
      No data container can have more than 1 parent. If it already has one,
      first unregister yourself from it:
    */
    if (fParent) {
      fParent -> removeChild(this);
    } // if
    fParent = aValue;
    /* If you have a parent now, register yourself with it: */
    if (fParent) {
      fParent -> addChild(this);
    } // if
  } // DataContainer::setParent

  void DataContainer::setSelf(void* aValue) {
    fSelf = (void**) aValue;
  } // DataContainer::setSelf

  void DataContainer::write(DataObject *aData) {
    fAdapter -> write(aData);
    if (fParent) {
      fParent -> write(aData);
    } // if
    QHashIterator<Associator*, Associator*> iter(fAssociators);
    while (iter.hasNext()) {
      iter.next();
      iter.value() -> write(aData, this);
    } // while
  } // DataContainer::write

} // GerQ
